#!/usr/bin/env python


from pyglet import window
from pyglet import clock
from pyglet import font
from pyglet import resource
from pyglet import sprite
from pyglet import graphics
from pyglet import gl
from pyglet import text


#Add Verlet code
from LinkedParticles import LinkedParticles

import random
import math
import time

def load_sprite(name, batch_draw, group_draw):
    img = resource.image(name)
    img.anchor_x = img.width*0.5
    img.anchor_y = img.height*0.5
    
    new_sprite = sprite.Sprite(img, batch=batch_draw, group=group_draw)
    
    return new_sprite

class BungeeJumper():
    def __init__ (self, batch_id, batch_group, **kwargs):        
        self.pos = [0,495]
        self.jump_x = random.randint(150,450)        
        self.patience = 100
        self.string_length = 0
        self.state = "walking" #walking,  waiting,  jumping,  finishing,  leaving, stop
        self.gone = False
        self.string = LinkedParticles(800,600,1.0)     
        self.accident = False
        self.satisfied = False
        self.speed = 0
        self.lowest_height = 500
        self.y_offset = 0
        
        self.img = resource.image("guy.png")
        self.img.anchor_x = self.img.width // 2
        
        self.img_angry = resource.image("guy_angry.png")
        self.img_angry.anchor_x = self.img_angry.width // 2
        
        self.img_accident = resource.image("guy_accident.png")
        self.img_accident.anchor_x = self.img_accident.width // 2
        
        self.images = [self.img, self.img_angry, self.img_accident]
        
        self.guy_img = sprite.Sprite(self.images[0], batch=batch_id, group=batch_group)
                
    def update(self):        
        #Walk until jump position is reached     
        if self.state == "walking":                             
            if self.pos[0] < self.jump_x:
                self.pos[0] = self.pos[0] + 2
            else:
                self.state = "waiting"
            
        #Wait to be chosen for jumping            
        elif self.state == "waiting":                        
            self.patience = self.patience - 1
            if self.patience <= 0:
                self.state = "leaving"
       
        #Let the Physics system move the guy, and check collisions
        elif self.state == "jumping":                       
            self.string.TimeStep()
            rope_begin = self.string.pos[0]        
            rope_end = self.string.pos[self.string.n_particles-1]        
            self.pos = rope_end[:]            
            
            #rotate guy image
            if (rope_begin[0] != rope_end[0]) and (rope_begin[1] != rope_end[1]):
                end_angle = math.atan2(rope_end[1] - rope_begin[1], rope_end[0] - rope_begin[0])
                end_angle = end_angle * 180.0 / math.pi
                self.guy_img.rotation = -(end_angle -90.0)
            else:
                self.guy_img.rotation = 180
            
            if (self.pos[1] < 450) and (self.speed < 10) and (self.string.total_correction < 10) and (abs(self.pos[0] - rope_begin[0]) < 3) :                
                self.state = "finishing"
               
        #Lift the jumper
        elif self.state == "finishing":            
            lifted = True
            for c in self.string.constraints:                
                if c["Length"] > 0:
                    c["Length"] = c["Length"] - 0.1
                    lifted = False
                else:
                    c["Length"] = 0
                
            self.string.TimeStep()

            rope_begin = self.string.pos[0]        
            rope_end = self.string.pos[self.string.n_particles-1]        
            self.pos = rope_end[:]
            
            #rotate guy image
            if (rope_begin[0] != rope_end[0]) and (rope_begin[1] != rope_end[1]):
                end_angle = math.atan2(rope_end[1] - rope_begin[1], rope_end[0] - rope_begin[0])
                end_angle = end_angle * 180.0 / math.pi
                self.guy_img.rotation = -(end_angle -90.0)
            else:
                self.guy_img.rotation = 180    
                                             
            if lifted:                
                self.guy_img.rotation = 0                
                self.state = "leaving"
                                   
        #Leave the bridge
        elif self.state == "leaving":
            if not self.satisfied and not self.accident:
                self.guy_img.image = self.images[1]
            self.pos[1] = 470
            if self.pos[0] > 0:
                self.pos[0] = self.pos[0] -2
            else:
                self.gone = True
        
        #Wait, I'm being clicked!
        elif self.state == "stop":
            pass
                
        #Update self speed and image position
        self.speed = abs(self.guy_img.x - self.pos[0]) + abs(self.guy_img.y - self.pos[1])
        self.guy_img.x = self.pos[0]
        self.guy_img.y = self.pos[1] 
        
    
    def jump (self, string_length, error):
        self.satisfied = True
        error = int(error * 100.0)
        half_error = int(error*0.75)        
        
        #Create string
        self.string_length = 10.0*string_length*40.0
        self.string.AddParticle(self.pos[0], self.pos[1])
        for i in range(0,8):
            if error < 0:
                self.string.AddParticle(self.pos[0] +random.randint(error, half_error), self.pos[1])
            elif error > 0:
                self.string.AddParticle(self.pos[0] +random.randint(half_error, error), self.pos[1])
            else:
                self.string.AddParticle(self.pos[0] +random.random()*0.01-0.005 , self.pos[1])
        self.string.AddParticle(self.pos[0], self.pos[1])
        
        for i in range(1,10):
            self.string.AddConstraint(i, i-1, string_length*40.0)
            
        self.string.mass[0] = 0
        self.state = "jumping"             
        
    def collision (self):
        self.accident = True
        self.guy_img.image = self.images[2]
        rope_end = self.string.pos[self.string.n_particles-1]        
        rope_end[0] = rope_end[0] + random.randint(-20,20)
        rope_end[1] = rope_end[1] + random.randint(-20,20)
        self.state = "finishing"
        
    def draw_line (self):
        gl.glColor3f(0.8, 0.6, 0.1)
        if self.string.n_constraints > 0:
            vertex_list = graphics.vertex_list(self.string.n_constraints*2, 'v2f' )
            vertices = []            
        
            for c in range(0, self.string.n_constraints):        
                posA = self.string.pos[self.string.constraints[c]["pA"]]        
                posB = self.string.pos[self.string.constraints[c]["pB"]]        
                vertices = vertices + [posA[0], posA[1], posB[0], posB[1]]
  
            vertex_list.vertices = vertices
            vertex_list.draw(gl.GL_LINES)
        
        
class JumpConfiguration():
    
    def __init__(self, batch_id, batch_group, **kwargs):
        self.x = 0
        self.state = "rope_length_start"
        self.time = 0
        self.precision = random.random() * 2.0 - 1.0
        self.inc_amount = 0.15 #should be higher if game difficulty rises
        self.current_inc = 1.0 
        self.ft = font.load('Arial', 18)
        self.ft = font.Text(self.ft, y=550) 
        self.length = 0
        self.error = 0
        img = resource.image("bungee_icon.png")
        img.anchor_x = img.width // 2
        img.anchor_y = img.height
        self.bungee_icon = sprite.Sprite(img, batch=batch_id, group=batch_group)
        
    def update(self):        
        if self.state == "precision_wheel":
            self.precision = self.precision + self.inc_amount*self.current_inc
            if self.precision > 1.0 and self.current_inc == 1.0:
                self.current_inc = -1.0
                self.precision = 1.0
            elif self.precision < -1.0 and self.current_inc == -1.0:
                self.current_inc = 1.0
                self.precision = -1.0
        
    def click(self, x_offset=0):        
        if self.state == "rope_length_start":
            self.bungee_icon.scale = 0.001
            self.bungee_icon.x = x_offset 
            self.bungee_icon.y = 500
            self.x = x_offset
            self.time = time.time()
            self.state = "rope_length_end"
        elif self.state == "rope_length_end":               
            self.length = min(1.25, time.time() - self.time)
            self.state = "precision_wheel" 
        elif self.state == "precision_wheel":
            self.error = self.precision
            self.state = "rope_length_start"
                
    def draw(self):  
        if self.state == "precision_wheel":
            self.bungee_icon.anchor_x = self.bungee_icon.width*0.5
            self.bungee_icon.rotation = -self.precision * 90.0
            
        elif self.state == "rope_length_end":
            length = (time.time() - self.time)
            if length > 1.25:
                length = 1.25
                return
            self.bungee_icon.scale = math.sqrt(length)*2.0
            
    
class TimedVisual():

    def __init__(self, image, gbatch, ggroup, pos, effect, modifiers):        
        self.effect = effect
        self.modifiers = modifiers
        self.image = load_sprite(image, gbatch, ggroup)
        self.image.x = pos[0]
        self.image.y = pos[1]
        self.lifetime = 50
        
    def update(self):        
        self.lifetime = self.lifetime-1
        
        if self.effect == "rise":
            self.image.y = self.image.y + 1
            self.image.opacity = int(255 - (50-self.lifetime)*4)
            
        if self.effect == "grow":
            if self.lifetime > 40:
                self.image.scale = self.image.scale + 0.05
                self.image.x = self.image.x - 1
                self.image.y = self.image.y - 1
        if self.effect == "shake":
            self.image.rotation = random.randint(-5, 5)

    def draw(self):
        pass
    
    
class Message():    
    def __init__(self, parent, label = "", img = None, batch="", group="", _x=0, _y=0):
        self.text = ""
        self.image = None
        if label != "":
            self.text = text.Label(label,
                                   font_name='Arial',
                                   font_size=18,
                                   x=250,
                                   y=200,
                                   anchor_x='center',
                                   anchor_y='center',
                                   width = 400,
                                   multiline=True,
                                   color = (255,255,255,255),
                                   bold = True)
        if img != None:
           self.img = load_sprite(img, batch, group)
           self.img.x = _x
           self.img.y = _y
            
        self.displaying = True        
        parent.push_handlers(self)
        
    def draw(self):
        if self.text != "":
            self.text.draw()        

    def on_key_press(self, symbol, modifiers):        
        self.displaying = False        
        
        
class BungeeMWindow(window.Window):
       
    def __init__(self, *args, **kwargs):

        #Let all of the standard stuff pass through
        window.Window.__init__(self, *args, **kwargs)       
                
        self.game_state = "Main menu"
        self.message = None
        
        #Add DATA directory to resource path
        resource.path.append('./data')
        resource.reindex()
        
        #Batch rendering
        self.batch_bkg = graphics.Batch()
        self.batch = graphics.Batch()        
        self.foreground = graphics.OrderedGroup(1)
        self.visualeffects = graphics.OrderedGroup(0)
           
        backgr = resource.image("background.png")
        self.backgr = sprite.Sprite(backgr, batch = self.batch_bkg)
                      
        self.jumpers = []
        self.chosen_guy = None
        self.jump_details = None
                
        self.timed_visuals = []
        
        #Economics
        self.benefits = 0
        self.happy_costumers = 1.0
        self.angry_costumers = 1.0
        self.tax_time = 0
        self.prob_new_costumers = 0.005
        img = resource.image("bungee_icon2.png")
        img.anchor_x = img.width*0.5
        img.anchor_y = img.height    
        self.success_meter_mark = sprite.Sprite(img, batch=self.batch, group=self.visualeffects)
        self.success_meter_mark.x = 700
        self.success_meter_mark.y = 220
        
        self.tax_mark = load_sprite("money.png", self.batch, self.foreground )        
        self.tax_mark.x = 635
        self.tax_mark.y = 75
        
        self.ft = font.load('Arial', 18)
        self.fps_text = font.Text(self.ft, y=10)

        self.benefits_text = font.Text(self.ft, x=650, y=360)
        
        self.main_menu()
        
    def reset_game(self):
        dummy_list = self.jumpers[:]        
        for jumper in dummy_list:
            self.jumpers.remove(jumper)
            del jumper
        dummy_list = self.timed_visuals[:]
        for visual in dummy_list:
            self.timed_visuals.remove(visual)
            del visual
        self.jumpers = []
        
        
        self.chosen_guy = None
        self.jump_details = None
        self.timed_visuals = []
        self.benefits = 0
        self.happy_costumers = 1.0
        self.angry_costumers = 1.0
        self.tax_time = 0
        self.prob_new_costumers = 0.005
        self.success_meter_mark.x = 700
        self.success_meter_mark.y = 220
        self.tax_mark.x = 635
        self.tax_mark.y = 75
        self.batch.draw()
        self.flip()
        
                                     
    def on_mouse_press(self, x, y, button, modifiers):      
        if self.jump_details == None:
            self.jump_details = JumpConfiguration(self.batch, self.foreground)
            
        if self.jump_details.state == "rope_length_start":            
            for jumper in self.jumpers:
                if jumper.state in ["jumping", "finishing", "leaving"]:
                    continue
                dist = abs(jumper.pos[0] - x + jumper.guy_img.width*0.5) + abs(jumper.pos[1] - y + jumper.guy_img.height*0.5)
                if (dist < 30) and (jumper.pos[0] > 100) and (jumper.pos[0] < 500):
                    self.chosen_guy = jumper                    
                    self.chosen_guy.state = "stop"
                    self.jump_details.click(jumper.pos[0])
                    break
        
        if self.jump_details.state == "precision_wheel":
            self.jump_details.click(0) 
            length = self.jump_details.length
            error = self.jump_details.error
            self.chosen_guy.jump(length, error)
            del self.jump_details
            self.jump_details = None
                    
                
    def on_mouse_release(self, x, y, button, modifiers):       
        if self.jump_details != None and self.jump_details.state == "rope_length_end":
            self.jump_details.click(0)
            
                
    def create_jumper(self):
        new_jumper = BungeeJumper(self.batch, self.foreground)
        self.jumpers.append(new_jumper)
        

    def update_jumpers(self):
        # Update Jumpers
        dummy_list = self.jumpers[:]
        currently_jumping = [] #who is  jumping, to check collisions
        for jumper in dummy_list:
            jumper.update()
            jumper.draw_line()                    
            
            if jumper.gone:
                if jumper.satisfied == True and jumper.accident == False:
                    self.happy_costumers = self.happy_costumers + 1
                else:
                    self.angry_costumers = self.angry_costumers + 1
                    self.benefits = self.benefits - 5
                self.jumpers.remove(jumper)
                del jumper
                continue
                
            #Check lowest height merits (each 50mts)            
            if (jumper.state == "jumping"):
                height_mark = int(math.floor(jumper.pos[1] / 50.0)) * 50                
                if height_mark < jumper.lowest_height:
                    jumper.lowest_height = height_mark
                    self.timed_visuals.append(TimedVisual("money.png", self.batch, self.visualeffects, jumper.pos, "rise", []))
                    self.benefits = self.benefits + height_mark / 10
                    if jumper.lowest_height == 50:
                        self.timed_visuals.append(TimedVisual("risky.png", self.batch, self.visualeffects, jumper.pos, "rise", []))
                        self.benefits = self.benefits + 250
                        
            #Check ground collision
            if (jumper.state == "jumping") and (jumper.pos[1] < 50):
                if jumper.speed > 5:
                    jumper.collision()                
                    self.timed_visuals.append(TimedVisual("clonk.png", self.batch, self.visualeffects, jumper.pos, "grow", []))
                    self.benefits = self.benefits - min(1000, int(jumper.speed)*50.0)
                    #scare clients away
                    for jumper in self.jumpers:
                        if jumper.state not in ["jumping", "finishing"]:
                            jumper.satisfied = False
                            jumper.state = "leaving"
                            
           
            #check if collides with another jumper: 
            if jumper.state == "jumping" or jumper.state == "finishing":
                for other_jumper in currently_jumping:
                    #check distances                    
                    if abs(other_jumper.pos[0] - jumper.pos[0]) + abs(other_jumper.pos[1] - jumper.pos[1]) < 30:                        
                        other_jumper.collision()
                        jumper.collision()
                        self.benefits = self.benefits - int(other_jumper.speed + jumper.speed)
                        self.timed_visuals.append(TimedVisual("bonk.png", self.batch, self.visualeffects, jumper.pos, "grow", []))
                currently_jumping.append(jumper)
                
        
    def update_timed_visuals(self):
        # Update timed visuals
        dummy_list = self.timed_visuals[:]

        for timed_visual in dummy_list:
            timed_visual.update()
            if timed_visual.lifetime <= 0:                                     
                self.timed_visuals.remove(timed_visual)
                del timed_visual
                
                
    def update_economics(self):
        success = self.happy_costumers - self.angry_costumers        
        
        #Sucess
        if success > 10:
            success = 10            
        if success < -10:
            success = -10
            self.end_game("Bad reputation")
        success_l = 4.0 * success        
        self.success_meter_mark.x = 700 + success_l
                
        #Taxes
        self.tax_mark.x = self.tax_mark.x + 0.1
        if self.tax_mark.x > 760:
            if self.benefits <= 0:
                self.end_game("Taxes not paid")
            else:
                self.timed_visuals.append(TimedVisual("taxes.png", self.batch, self.visualeffects, [700, 75], "rise", []))
                self.benefits = self.benefits * 0.86
                self.tax_mark.x = 635
                self.tax_mark.y = 75
                if self.prob_new_costumers < 0.05:
                    self.prob_new_costumers = self.prob_new_costumers + 0.0025
                
                
    def main_menu(self):        
        self.game_state = "Main menu"
        message = '''Keep your clients happy and pay your taxes to keep your business open!
        
        Press any key start game...'''
        self.message = Message(self, img="instructions.png", _x=400,_y=300, batch=self.batch, group=self.foreground)
        
        
    def end_game(self, problem):        
        self.game_state = "Game end"
        message = ""
        if problem == "Bad reputation":
            message = '''GAME OVER!
            
    Too many angry clients spreading rumors are not good for your business! Keep them happier next time!
            
    Press any key to continue...'''
        if problem == "Taxes not paid":
            message = '''GAME OVER!
            
    Sorry, but you don't have enough money for paying taxes. You must end your activity!
            
    Press any key to continue...'''
        self.message = Message(self, label=message, img=None)
        
        
    def main_loop(self):
            while not self.has_exit:                            
                self.clear()
                self.batch_bkg.draw()
                self.update_timed_visuals()                
                clock.set_fps_limit(30)                     
                self.dispatch_events()   
                
                #Manage game states and floating texts...                
                if self.message != None:                    
                    if self.message.displaying == True:
                        self.message.draw()                                            
                        self.batch.draw()
                        self.flip()
                    if self.message.displaying == False:                                                
                        self.pop_handlers()
                        self.message = None
                        if self.game_state == "Main menu":                            
                            self.game_state = "Game"
                            self.create_jumper()
                        elif self.game_state == "Game end":
                            self.reset_game()
                            self.batch.draw()
                            self.flip()
                            self.main_menu()                            
                    continue                                                       
                
                if random.random() < self.prob_new_costumers:
                    self.create_jumper()
                                            
                self.update_jumpers()                
                       
                # Update interface
                if self.jump_details != None:
                    self.jump_details.update()
                    self.jump_details.draw()
                   
                self.update_economics()
                
                self.batch.draw()
                                    
                #Tick the clock
                clock.tick()
                
                #Gets fps and draw it
                self.benefits_text.text = ("%d") % (self.benefits)
                self.benefits_text.draw()
                self.fps_text.text = ("fps: %d") % (clock.get_fps())
                self.fps_text.draw()
               
                self.flip()
